# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 18 Feb, 2018
#
from tornado_battery.queue import register_queue_options, QueueConnectorError
from tornado_battery.queue import JSONQueue, JSONQueueError
import pytest

pytestmark = pytest.mark.asyncio
register_queue_options("test", "amqp://guest:guest@127.0.0.1:5672/%2F")
register_queue_options("vanilla", "amqp://guest:guest@127.0.0.1:5672/%2F")


@pytest.fixture
async def queue():
    pika = JSONQueue.instance("test")
    await pika.connect()
    return pika


@pytest.fixture
async def vanilla_queue():
    pika = JSONQueue.instance("vanilla")
    await pika.connect()
    return pika


async def test_publish_consume(queue):
    await queue.setup(queue="test_queue", exchange="test_exchange")
    await queue.publish(dict(value=1984))

    data, msg = await queue.consume()
    msg.ack()
    assert data == dict(value=1984)


async def test_no_setup_publish(vanilla_queue):
    with pytest.raises(JSONQueueError, match="^no active exchange set$"):
        await vanilla_queue.publish(dict(value=1984))


async def test_no_setup_consume(vanilla_queue):
    with pytest.raises(JSONQueueError, match="^no active queue set$"):
        await vanilla_queue.consume()


async def test_no_connection():
    match = r"^no connection of noconn found$"
    with pytest.raises(QueueConnectorError, match=match):
        JSONQueue.instance("noconn").connection()


async def test_option_name():
    from tornado_battery.queue import option_name

    assert option_name("master", "uri") == "queue-master-uri"
    assert option_name("slave", "uri") == "queue-slave-uri"



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 16 Feb, 2018
#
import time
from tornado_battery.cassandra import (
    register_cassandra_options, with_cassandra)
from tornado_battery.cassandra import (
    CassandraConnector, CassandraConnectorError)
import pytest

pytestmark = pytest.mark.asyncio
register_cassandra_options("test")


@pytest.fixture
async def cassandra():
    cassandra = CassandraConnector.instance("test")
    await cassandra.connect()
    session = cassandra.connection()
    cql_keyspace_create = """
        CREATE KEYSPACE IF NOT EXISTS testtmp WITH
        replication = {'class': 'SimpleStrategy',
        'replication_factor' : 1}
    """
    q = await session.prepare_future(cql_keyspace_create)
    await session.execute_future(q)
    q = await session.prepare_future('USE testtmp')
    await session.execute_future(q)
    cql_tb_create = ('CREATE TABLE tmp_tornado_battery_test ('
                     'specid text PRIMARY KEY,'
                     'foo text,'
                     'bar int'
                     ')')
    cql_tb_drop = 'DROP TABLE IF EXISTS tmp_tornado_battery_test'
    q = await session.prepare_future(cql_tb_drop)
    await session.execute_future(q)
    q = await session.prepare_future(cql_tb_create)
    await session.execute_future(q)
    return cassandra


async def test_cql(cassandra):
    now = int(time.time())
    cql = """
        INSERT INTO tmp_tornado_battery_test (specid, foo, bar)
        VALUES ('test', 'foo0', ?)
    """
    session = cassandra.connection()
    query = await session.prepare_future(cql)
    await session.execute_future(query, (now, ))
    cql = """
        SELECT foo, bar FROM tmp_tornado_battery_test WHERE specid='test';
    """
    query = await session.prepare_future(cql)
    values = await session.execute_future(query)
    assert(values[0][0] == 'foo0')
    assert(values[0][1] == now)


async def test_decorator(cassandra):
    now = int(time.time())

    @with_cassandra(name="test")
    async def _read(now, cassandra=None):
        cql = """
            INSERT INTO tmp_tornado_battery_test (specid, foo, bar)
            VALUES ('test', 'foo0', ?)
        """
        query = await cassandra.prepare_future(cql)
        await cassandra.execute_future(query, (now, ))
        cql = """
            SELECT foo, bar FROM tmp_tornado_battery_test WHERE specid='test';
        """
        query = await cassandra.prepare_future(cql)
        values = await cassandra.execute_future(query)
        return values
    values = await _read(now)
    assert(values[0][0] == 'foo0')
    assert(values[0][1] == now)


async def test_decorator_duplicated(cassandra):

    @with_cassandra(name="test")
    async def _read(cassandra):
        cql = """
            SELECT foo, bar FROM tmp_tornado_battery_test WHERE specid='test';
        """
        query = await cassandra.prepare_future(cql)
        values = await cassandra.execute_future(query)
        return values

    with pytest.raises(CassandraConnectorError):
        await _read(cassandra=None)


async def test_invalid_connection_scheme():
    from tornado.options import options
    options.cassandra_test_uri = "test://"
    match = r" is not a cassandra connection scheme$"
    with pytest.raises(CassandraConnectorError, match=match):
        await CassandraConnector.instance("test").connect()


async def test_option_name():
    from tornado_battery.cassandra import option_name

    assert option_name("test", "points") == "cassandra-test-points"



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 16 Feb, 2018
#
from tornado_battery.extra import try_get_value
import pytest


@pytest.fixture
def data():

    return {
        "1": {
            "A": {
                "a": "value"
            },
            "B": "Bvalue"
        }
    }


def test_last_node(data):
    assert try_get_value(data, "1.A.a", None) == "value"


def test_last_non_exists(data):
    assert try_get_value(data, "1.A.b", "n/a") == "n/a"


def test_middle_non_exists(data):
    assert try_get_value(data, "1.c.a", "n/a") == "n/a"


def test_middle_node(data):
    assert try_get_value(data, "1.B", "n/a") == "Bvalue"


def test_empty(data):
    assert try_get_value(data, "", "n/a") == "n/a"


def test_non_dict():
    assert try_get_value(1, "", "n/a") == "n/a"



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 16 Feb, 2018
#
from tornado_battery.exception import GeneralException
import pytest


class NonFormatException(GeneralException):
    error_format = "a non-format message"


class SimpleFormatException(GeneralException):
    error_format = "an exception: %(reason)s"


class NoMessageException(GeneralException):
    pass


class ErrorCodeException(GeneralException):
    error_code = 40001


def test_non_format_exception():
    match = r"^a non-format message$"
    with pytest.raises(NonFormatException, match=match):
        raise NonFormatException()


def test_format_exception():
    match = "^an exception: something wrong$"
    with pytest.raises(SimpleFormatException, match=match):
        raise SimpleFormatException(reason="something wrong")


def test_error_format_exception():
    match = ("^cannot format exception message: "
             "'an exception: %\(reason\)s'$")
    with pytest.raises(SimpleFormatException, match=match):
        raise SimpleFormatException(other="something wrong")


def test_no_message_exception():
    match = "application exception occurred: something wrong"
    with pytest.raises(NoMessageException, match=match):
        raise NoMessageException(reason="something wrong")


def test_error_code_exception():
    match = "application exception occurred: something wrong"
    with pytest.raises(ErrorCodeException, match=match) as ei:
        raise ErrorCodeException(reason="something wrong", error_code=4001)
    assert ei.value.error_code == 4001


def test_error_code_set():
    match = "application exception occurred: something wrong"
    with pytest.raises(NoMessageException, match=match) as ei:
        raise NoMessageException(reason="something wrong", error_code=4002)
    assert ei.value.error_code == 4002



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @  8 Feb, 2018
#
from tornado.options import define, options
from tornado.options import parse_config_file

import logging
import logging.config
import tornado


define("debug", group="main", default=False, help="enable debug")
define("config", help="path to config file", group='main',
       callback=lambda path: parse_config_file(path, final=False))
define('logging', default='info', group='main',
       metavar='debug|info|warn|error|none',
       help="logging level")
define("logging-config", default='', group='main',
       help="path to logging config file")


DEFAULT_LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'colorlog': {
            '()': 'colorlog.ColoredFormatter',
            'format': ('%(log_color)s<%(process)d>[%(levelname)1.1s '
                       '%(asctime)s %(module)s:%(lineno)d] '
                       '%(message)s%(reset)s'),
            'datefmt': '%y%m%d %H:%M:%S',
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'colorlog',
        },
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': True
        },
        'app.biz': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': False
        },
    }
}


class CommandMixin:

    def bail(self, future, io_loop):
        io_loop.stop()
        exc = future.exception()
        if exc:
            raise exc

    def enable_logging(self):
        if options.debug:
            level = 'DEBUG'
        else:
            level = options.logging.upper()
        logging.config.dictConfig(DEFAULT_LOGGING_CONFIG)
        if options.logging_config:
            logging.config.fileConfig(options.logging_config,
                                      disable_existing_loggers=False)
        logger = logging.getLogger()
        logger.setLevel(getattr(logging, level))

    def setup(self):
        pass

    def before_run(self, io_loop):
        pass

    def __call__(self):
        from tornado.options import parse_command_line

        io_loop = tornado.ioloop.IOLoop.current()
        self.setup()
        parse_command_line()
        self.enable_logging()
        self.before_run(io_loop)
        io_loop.start()



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 16 Feb, 2018
#
from tornado_battery.redis import register_redis_options, with_redis
from tornado_battery.redis import RedisConnector, RedisConnectorError
import pytest

pytestmark = pytest.mark.asyncio
register_redis_options("test", "redis://127.0.0.1/0")


@pytest.fixture
async def redis():
    redis = RedisConnector.instance("test")
    await redis.connect()
    return redis


async def test_set_command(redis):
    async with redis.connection() as db:
        value = await db.execute("set", "redis_value", "1984")
    assert value == "OK"


async def test_get_command(redis):
    async with redis.connection() as db:
        await db.execute("set", "redis_value", "1984")
        value = await db.execute("get", "redis_value")
    assert value == "1984"


async def test_decorator(redis):

    @with_redis(name="test")
    async def _read(redis):
        value = await redis.execute("set", "redis_decorator_value", "1983")
        return value
    assert (await _read()) == "OK"


async def test_decorator_duplicated(redis):

    @with_redis(name="test")
    async def _read(redis):
        value = await redis.execute("set", "redis_decorator_value", "1983")
        return value
    with pytest.raises(RedisConnectorError):
        await _read(redis=None)


async def test_no_connection():
    match = r"^no connection of noconn found$"
    with pytest.raises(RedisConnectorError, match=match):
        RedisConnector.instance("noconn").connection()


async def test_invalid_connection_scheme():
    from tornado.options import options
    options.redis_test_uri = "test://"
    match = r" is not a redis connection scheme$"
    with pytest.raises(RedisConnectorError, match=match):
        await RedisConnector.instance("test").connect()


async def test_option_name():
    from tornado_battery.redis import option_name

    assert option_name("master", "uri") == "redis-master-uri"
    assert option_name("slave", "uri") == "redis-slave-uri"



# -*- coding: utf-8 -*-

from .exception import ServerException
from .pattern import NamedSingletonMixin
from tornado.options import define, options
from urllib.parse import urlparse

import asyncio
import aiomysql
import functools
import logging

LOG = logging.getLogger('tornado.application')


def query_as_dict(query: str) -> dict:
    result = dict()
    if not query:
        return result
    item_list = query.split('&')
    for item in item_list:
        key, value = item.split('=')
        result[key] = value
    return result


class MysqlConnectorError(ServerException):
    pass


class MysqlConnector(NamedSingletonMixin):

    def __init__(self, name: str):
        self.name = name

    def connection(self):
        if not hasattr(self, '_connections') or not self._connections:
            raise MysqlConnectorError("no connection of %s found" % self.name)
        return self._connections.acquire()

    def setup_options(self):
        name = self.name
        opts = options.group_dict('%s database' % name)
        uri = opts[option_name(name, "uri")]
        r = urlparse(uri)
        q = query_as_dict(r.query)
        if r.scheme.lower() != 'mysql':
            raise MysqlConnectorError('%s is not a mysql connection scheme' % uri)
        self._host = r.hostname or 'localhost'
        self._port = r.port or 3306
        self._user = r.username
        self._password = r.password
        self._db = r.path.lstrip('/') or r.username
        self._charset = q.get('charset') or 'utf8mb4'
        fmt = ('host={host} port={port} dbname={db} '
               'user={user} password={pw} charset={charset}')
        dsn = fmt.format(host=self._host,
                         port=self._port,
                         user=self._user,
                         pw=self._password,
                         db=self._db,
                         charset=self._charset)
        self._connection_string = dsn
        self._num_connections = opts[option_name(name, "num-connections")]
        self._pool_recycle = opts[option_name(name, "pool-recycle")]

    async def connect(self, autocommit=True, event_loop=None):
        self.setup_options()
        LOG.info('connecting mysql [%s] %s' %
                 (self.name, self._connection_string))
        if event_loop is None:
            event_loop = asyncio.get_event_loop()
        self._connections = await aiomysql.create_pool(
            host=self._host,
            port=self._port,
            user=self._user,
            password=self._password,
            db=self._db,
            charset=self._charset,
            minsize=int(self._num_connections[0]),
            maxsize=int(self._num_connections[-1]),
            autocommit=autocommit,
            pool_recycle=self._pool_recycle,
            loop=event_loop,
        )
        return self._connections


def option_name(instance: str, option: str) -> str:
    return 'mysql-%s-%s' % (instance, option)


def register_mysql_options(instance: str='master', default_uri: str='mysql:///'):
    define(option_name(instance, "uri"),
           default=default_uri,
           group='%s database' % instance,
           help="mysql connection uri for %s" % instance)
    define(option_name(instance, 'num-connections'), multiple=True,
           default=[1, 4],
           group='%s database' % instance,
           help='connection pool size for %s ' % instance)
    define(option_name(instance, 'pool-recycle'),
           default=30,
           group='%s database' % instance,
           help='pool recycle timeout for %s' % instance)


def with_mysql(name: str):

    def wrapper(function):

        @functools.wraps(function)
        async def f(*args, **kwargs):
            async with MysqlConnector.instance(name).connection() as db:
                LOG.debug("mysql connection acquired.")
                if "db" in kwargs:
                    raise MysqlConnectorError(
                        "duplicated database argument for database %s" % name)
                kwargs.update({"db": db})
                retval = await function(*args, **kwargs)
                return retval
        return f

    return wrapper


def connect_mysql(name: str, autocommit: bool=True):
    async def _connect(event_loop=None):
        return await MysqlConnector.instance(name).connect(autocommit, event_loop)
    return _connect



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 17 Feb, 2018
#
from .exception import ServerException
from .pattern import NamedSingletonMixin
from aio_pika import connect_robust, Message
from tornado.options import define, options
from typing import Any
from ujson import dumps as json_encode, loads as json_decode

import asyncio
import logging

LOG = logging.getLogger('tornado.application')


class QueueConnectorError(ServerException):
    pass


class QueueConnector(NamedSingletonMixin):

    def __init__(self, name):
        self.name = name

    async def connect(self, event_loop=None):
        if event_loop is None:
            event_loop = asyncio.get_event_loop()
        name = self.name
        opts = options.group_dict('%s queue' % name)
        uri = opts[option_name(name, "uri")]
        LOG.info('connecting amqp [%s] %s' % (self.name, uri))
        self._connection = await connect_robust(uri, loop=event_loop)

    def connection(self):
        if not hasattr(self, '_connection') or not self._connection:
            raise QueueConnectorError("no connection of %s found" % self.name)
        return self._connection


class JSONQueueError(ServerException):
    pass


class JSONQueue(QueueConnector):

    async def setup(self, queue: str=None, exchange: str='', routing_key: str='',
                    durable: bool=False):
        channel = await self.connection().channel()
        exchange = await channel.declare_exchange(exchange)
        if queue:
            queue = await channel.declare_queue(queue, auto_delete=True,
                                                durable=durable)
            await queue.bind(exchange, routing_key)
            self.active_queue = queue
        self.routing_key = routing_key
        self.active_exchange = exchange

    async def publish(self, content: Any):
        if not hasattr(self, "active_exchange"):
            raise JSONQueueError("no active exchange set")

        message = Message(bytes(json_encode(content), "utf-8"),
                          content_type="application/json")
        await self.active_exchange.publish(message, self.routing_key)

    async def consume(self):
        if not hasattr(self, "active_queue"):
            raise JSONQueueError("no active queue set")

        msg = await self.active_queue.get()
        if msg.content_type != 'application/json':
            msg.reject(requeue=False)
            return None
        try:
            data = json_decode(msg.body)
        except ValueError as e:
            msg.reject(requeue=False)
            return None
        return data, msg


def option_name(instance: str, option: str) -> str:
    return 'queue-%s-%s' % (instance, option)


def register_queue_options(instance: str='master', default_uri: str='amqp:///'):
    define(option_name(instance, "uri"),
           default=default_uri,
           group='%s queue' % instance,
           help="queue connection uri for %s" % instance)



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @  8 Feb, 2018
#
from typing import Any, Dict


def try_get_value(d: Dict[str, Any], path: str, null_value: Any=None):

    current, nodes = d, path.split('.')

    while nodes:
        if not isinstance(current, dict):
            return null_value
        key = nodes.pop(0)
        if key not in current:
            return null_value
        current = current[key]

    return current



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 17 Feb, 2018
#
from tornado_battery.postgres import register_postgres_options, with_postgres
from tornado_battery.postgres import PostgresConnector, PostgresConnectorError
import pytest

pytestmark = pytest.mark.asyncio
register_postgres_options("test", "postgres://test:test@127.0.0.1/test")


@pytest.fixture
async def postgres():
    db = PostgresConnector.instance("test")
    await db.connect()
    return db


async def test_select(postgres):

    @with_postgres(name="test")
    async def _select(db):
        async with db.cursor() as cursor:
            await cursor.execute("SELECT 1")
            value = await cursor.fetchone()
            return value

    assert (await _select()) == (1,)


async def test_decorator_duplicated(postgres):

    @with_postgres(name="test")
    async def _select(db):
        async with db.cursor() as cursor:
            await cursor.execute("SELECT 1")
            value = await cursor.fetchone()
            return value
    with pytest.raises(PostgresConnectorError):
        await _select(db=None)


async def test_no_connection():
    match = r"^no connection of noconn found$"
    with pytest.raises(PostgresConnectorError, match=match):
        PostgresConnector.instance("noconn").connection()


async def test_invalid_connection_scheme():
    from tornado.options import options
    options.postgres_test_uri = "test://"
    match = r" is not a postgres connection scheme$"
    with pytest.raises(PostgresConnectorError, match=match):
        await PostgresConnector.instance("test").connect()


async def test_option_name():
    from tornado_battery.postgres import option_name

    assert option_name("master", "uri") == "postgres-master-uri"
    assert option_name("slave", "uri") == "postgres-slave-uri"



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @  8 Feb, 2018
#


class GeneralException(Exception):
    error_format = "application exception occurred: %(reason)s"

    def __init__(self, message: str=None, **kwargs):
        self.kwargs = kwargs

        if not message:
            try:
                message = self.error_format % self.kwargs
            except KeyError:
                message = ("cannot format exception message: '%s'" %
                           self.error_format)

        if "error_code" in kwargs:
            self.error_code = kwargs["error_code"]

        self.message = message
        super().__init__(message)


class ClientException(GeneralException):
    error_format = "%(reason)s"
    error_code = 400000


class ServerException(GeneralException):
    error_format = "%(reason)s"
    error_code = 500000



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 28 Apr, 2018
#
from asyncio import wait as async_wait
from tornado_battery.redis import register_redis_options, with_redis
from tornado_battery.redis import RedisConnector, RedisConnectorError
from tornado_battery.locks import throttle, ThrottleExceeded
import pytest

pytestmark = pytest.mark.asyncio
register_redis_options("locks", "redis://127.0.0.1/0")


@pytest.fixture
async def redis():
    redis = RedisConnector.instance("locks")
    await redis.connect()
    yield redis


@with_redis(name="locks")
async def clear(key, redis):
    await redis.execute("DEL", f'throttle_{key}')


async def test_throttle_serial(redis):

    @with_redis(name="locks")
    async def _read(redis):
        async with throttle(redis, "TEST_LOCK_1", 2, 5) as value:
            return value

    await clear("TEST_LOCK_1")
    value = await _read()
    value = await _read()
    assert value == 2


async def test_throttle_concurrent(redis):

    @with_redis(name="locks")
    async def _read(redis):
        async with throttle(redis, "TEST_LOCK_2", 2, 5) as value:
            return value

    await clear("TEST_LOCK_2")
    tasks, _ = await async_wait([_read(), _read()])
    values = list(map(lambda x: x.result(), tasks))


async def test_throttle_exceeded(redis):

    @with_redis(name="locks")
    async def _read(redis):
        async with throttle(redis, "TEST_LOCK_3", 1, 5) as value:
            return value

    await clear("TEST_LOCK_3")
    with pytest.raises(ThrottleExceeded):
        tasks, _ = await async_wait([_read(), _read()])
        values = list(map(lambda x: x.result(), tasks))


async def test_throttle_with_release_serial(redis):

    @with_redis(name="locks")
    async def _read(redis):
        async with throttle(redis, "TEST_LOCK_4", 1, 5, release=True) as value:
            return value

    await clear("TEST_LOCK_4")
    value = await _read()
    value = await _read()
    assert value == 1


async def test_throttle_with_release_concurrent(redis):

    @with_redis(name="locks")
    async def _read(redis):
        async with throttle(redis, "TEST_LOCK_5", 1, 5, release=True) as value:
            return value

    await clear("TEST_LOCK_5")
    with pytest.raises(ThrottleExceeded):
        tasks, _ = await async_wait([_read(), _read()])
        values = list(map(lambda x: x.result(), tasks))



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 12 Feb, 2018
#
from .exception import ServerException
from .pattern import NamedSingletonMixin
from tornado.options import define, options
from urllib.parse import urlparse

import asyncio
import functools
import logging
from cassandra.cluster import Cluster
from cassandra.policies import RoundRobinPolicy
from aiocassandra import aiosession


LOG = logging.getLogger('tornado.application')


def query_as_dict(query: str) -> dict:
    result = dict()
    if not query:
        return result
    item_list = query.split('&')
    for item in item_list:
        key, value = item.split('=')
        result[key] = value
    return result


class CassandraConnectorError(ServerException):
    pass


class CassandraConnector(NamedSingletonMixin):

    def __init__(self, name: str):  # NOQA
        self.name = name

    def connection(self):
        if not hasattr(self, '_session') or not self._session:
            raise CassandraConnectorError(
                "no session of %s found" % self.name)
        return self._session

    async def connect(self, load_balancing_policy=None):
        name = self.name
        opts = options.group_dict('%s cassandra' % name)
        uri = opts[option_name(name, "uri")]
        r = urlparse(uri)
        q = query_as_dict(r.query)
        if r.scheme.lower() != 'cassandra':
            raise CassandraConnectorError('%s is not a cassandra connection scheme' % uri)
        self._hosts = r.hostname.split(',') if r.hostname else ['127.0.0.1']
        self._port = r.port or 9042
        self._keyspace = None
        keyspace = r.path.lstrip('/')
        if keyspace:
            self._keyspace = keyspace
        executor_threads = opts[option_name(name, 'executor-threads')]
        _load_balancing_policy = load_balancing_policy or RoundRobinPolicy()
        LOG.info('connecting cassandra [%s] %s' % (self.name, uri))

        # TODO 选取合适的loadbalancingpolicy
        # Cluster.__init__ called with contact_points specified
        # should specify a load-balancing policy
        # http://datastax.github.io/python-driver/_modules/cassandra/cluster.html#Cluster # NOQA
        # RoundRobinPolicy:
        # http://datastax.github.io/python-driver/_modules/cassandra/policies.html#RoundRobinPolicy # NOQA
        cluster = Cluster(contact_points=self._hosts,
                          port=self._port, executor_threads=executor_threads,
                          load_balancing_policy=_load_balancing_policy)
        self._session = cluster.connect(keyspace=self._keyspace)
        aiosession(self._session)


def option_name(instance: str, option: str) -> str:
    return 'cassandra-%s-%s' % (instance, option)


def register_cassandra_options(
        instance: str='default',
        default_uri: str='cassandra:///',
        executor_threads: int=2):
    define(option_name(instance, "uri"),
           default=default_uri,
           group='%s cassandra' % instance,
           help='cassandra contact uri %s' % instance)
    define(option_name(instance, 'executor-threads'),
           default=executor_threads,
           group='%s cassandra' % instance,
           help='num of cassandra executor_threads for %s ' % instance)


def with_cassandra(name: str):

    def wrapper(function):

        @functools.wraps(function)
        async def f(*args, **kwargs):
            cassandra = CassandraConnector.instance(name).connection()
            if 'cassandra' in kwargs:
                raise CassandraConnectorError(
                    'duplicated argument for cassandra %s' % name)
            kwargs.update({'cassandra': cassandra})
            retval = await function(*args, **kwargs)
            return retval
        return f

    return wrapper


def connect_cassandra(name: str):
    return CassandraConnector.instance(name).connect



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 17 Feb, 2018
#
from tornado_battery.mysql import register_mysql_options, with_mysql
from tornado_battery.mysql import MysqlConnector, MysqlConnectorError
from pymysql.err import IntegrityError
import pytest
import logging

pytestmark = pytest.mark.asyncio
register_mysql_options("test", "mysql://root@127.0.0.1:3306/test")
logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger("test_mysql")

@pytest.fixture
async def mysql():
    db = MysqlConnector.instance("test")
    await db.connect()
    LOG.info("connect success!")

    @with_mysql(name="test")
    async def _init(db):
        async with db.cursor() as cursor:
            await cursor.execute("CREATE TABLE IF NOT EXISTS "
                                 "test(id SERIAL PRIMARY KEY, x INT, y INT, "
                                 "UNIQUE KEY (x, y));")
            await cursor.execute("TRUNCATE TABLE test")
            await cursor.execute("INSERT INTO test(x, y) VALUES(1, 1)")
            await cursor.execute("INSERT INTO test(x, y) VALUES(1, 2)")
    await _init()

    return db

async def test_select(mysql):

    @with_mysql(name="test")
    async def _select(db):
        async with db.cursor() as cursor:
            await cursor.execute("SELECT 1")
            value = await cursor.fetchone()
            return value

    assert (await _select()) == (1,)


async def test_error_transaction(mysql):

    @with_mysql(name="test")
    async def _insert(db):
        async with db.cursor() as cursor:
            await cursor.execute("BEGIN")
            await cursor.execute("INSERT INTO test(x, y) VALUES(2, 2)")
            await cursor.execute("INSERT INTO test(x, y) VALUES(1, 1)")
            await cursor.execute("COMMIT")

    @with_mysql(name="test")
    async def _select(db):
        async with db.cursor() as cursor:
            await cursor.execute("SELECT * FROM test WHERE x = 2 AND y = 2")
            value = await cursor.fetchone()
        return value

    with pytest.raises(IntegrityError):
        await _insert()

    assert (await _select()) == None

async def test_rollback_transaction(mysql):

    @with_mysql(name="test")
    async def _insert(db):
        async with db.cursor() as cursor:
            await cursor.execute("BEGIN")
            await cursor.execute("INSERT INTO test(x, y) VALUES(3, 3)")
            await cursor.execute("INSERT INTO test(x, y) VALUES(3, 4)")
            await cursor.execute("ROLLBACK")

    @with_mysql(name="test")
    async def _select(db):
        async with db.cursor() as cursor:
            await cursor.execute("SELECT * FROM test WHERE x = 2 AND y = 2")
            value = await cursor.fetchone()
        return value

    await _insert()
    assert (await _select()) == None

async def test_commit_transaction(mysql):

    @with_mysql(name="test")
    async def _insert(db):
        async with db.cursor() as cursor:
            await cursor.execute("BEGIN")
            await cursor.execute("INSERT INTO test(x, y) VALUES(4, 3)")
            await cursor.execute("INSERT INTO test(x, y) VALUES(4, 4)")
            await cursor.execute("COMMIT")

    @with_mysql(name="test")
    async def _select(db):
        async with db.cursor() as cursor:
            await cursor.execute("SELECT * FROM test WHERE x = 3")
            value = await cursor.fetchone()
        return value

    await _insert()
    print (await _select())

async def test_decorator_duplicated(mysql):

    @with_mysql(name="test")
    async def _select(db):
        async with db.cursor() as cursor:
            await cursor.execute("SELECT 1")
            value = await cursor.fetchone()
            return value
    with pytest.raises(MysqlConnectorError):
        await _select(db=None)

async def test_no_connection():
    match = r"^no connection of noconn found$"
    with pytest.raises(MysqlConnectorError, match=match):
        MysqlConnector.instance("noconn").connection()


async def test_invalid_connection_scheme():
    from tornado.options import options
    options.mysql_test_uri = "test://"
    match = r" is not a mysql connection scheme$"
    with pytest.raises(MysqlConnectorError, match=match):
        await MysqlConnector.instance("test").connect()


async def test_option_name():
    from tornado_battery.mysql import option_name

    assert option_name("master", "uri") == "mysql-master-uri"
    assert option_name("slave", "uri") == "mysql-slave-uri"



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 16 Feb, 2018
#
from tornado_battery.pattern import SingletonMixin, NamedSingletonMixin


class Single1(SingletonMixin):
    pass


class Single2(SingletonMixin):
    pass


class NamedSingle1(NamedSingletonMixin):

    def __init__(self, name):
        self.name = name


class NamedSingle2(NamedSingletonMixin):

    def __init__(self, name):
        self.name = name


def test_singleton():
    Single2.instance()
    instance_id = id(Single1.instance())
    ids = [id(Single1.instance()) == instance_id for i in range(0, 100)]
    assert all(ids)
    assert id(Single1.instance()) != id(Single2.instance())


def test_named_singleton():
    NamedSingle2.instance("master")
    NamedSingle2.instance("slave")
    master_id = id(NamedSingle1.instance("master"))
    master_ids = [id(NamedSingle1.instance("master")) == master_id
                  for i in range(0, 100)]
    assert all(master_ids)

    slave_id = id(NamedSingle1.instance("slave"))
    slave_ids = [id(NamedSingle1.instance("slave")) == slave_id
                 for i in range(0, 100)]
    assert all(slave_ids)

    assert master_id != slave_id
    assert id(NamedSingle1.instance("master")) != id(NamedSingle2.instance("master"))
    assert id(NamedSingle1.instance("slave")) != id(NamedSingle2.instance("slave"))



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @  8 Feb, 2018
#

import logging
import threading
LOG = logging.getLogger("tornado.application")


class SingletonMixin:

    __instance_lock = threading.Lock()

    @classmethod
    def instance(cls):
        if not hasattr(cls, "__instance"):
            with cls.__instance_lock:
                if not hasattr(cls, "__instance"):
                    LOG.debug("create a new instance for '%s'" % cls)
                    setattr(cls, "__instance", cls())

        return getattr(cls, "__instance")


class NamedSingletonMixin:

    __instance_lock = threading.Lock()

    @classmethod
    def instance(cls, name: str):
        if not hasattr(cls, "__instances"):
            with cls.__instance_lock:
                if not hasattr(cls, "__instances"):
                    LOG.debug("create a instances storage for '%s'" % cls)
                    setattr(cls, "__instances", dict())

        instances = getattr(cls, "__instances")
        if name not in instances:
            with cls.__instance_lock:
                instances[name] = cls(name=name)
        return instances[name]



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 28 Apr, 2018
#
"""
A set of utilities which leveraging redis to do synchronization easily
for common business logic.
"""
from asyncio_extras import async_contextmanager
import logging


LOG = logging.getLogger("tornado.application")


class ThrottleExceeded(Exception):

    def __init__(self, name, amount):
        self.name = name
        self.amount = amount


@async_contextmanager
async def throttle(redis, name, amount, expire=None, release=False):
    """
    Example:

    A constraint prevent any user from bidding more than 15
    times every 600 seconds.

    Code:

        async with throttle(redis, "BIDS_%s" % user_id, 15, 600) as value:
            # doing something

    Notes:

        The value of the throttle will be increased every time
        entering the context and decreased every time leaving
        the context. NOTICE, any exception inside the context
        will cause the throttle decreased
.
    """
    redis_key = "throttle_%s" % name
    value = await redis.execute("INCR", redis_key)
    if expire:
        await redis.expire(redis_key, expire)
    if value <= amount:
        try:
            yield value
            if release:
                await redis.execute("DECR", redis_key)
        except Exception as e:
            LOG.warn("throttle %s decreased due to exception", name)
            await redis.execute("DECR", redis_key)
            raise e
    else:
        LOG.info("throttle %s exceeded %s", name, amount)
        await redis.execute("DECR", redis_key)
        raise ThrottleExceeded(name, amount)



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @  8 Feb, 2018
#
from os.path import dirname, realpath, join as path_join
from setuptools import setup, find_packages

package = 'tornado_battery'
version = '0.5.20'


def valid_requirement(line):
    if not line:
        return False
    else:
        ch = line[0]
        return ch not in ('#', '-')


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    root = dirname(realpath(__file__))
    line_iter = (line.strip() for line in open(path_join(root, filename)))
    return [line for line in line_iter if valid_requirement(line)]


setup(
    name=package,
    version=version,
    description='utilities that help write tornado apps quickly and easily',
    url='https://github.com/jianingy/tornado_battery',
    setup_requires=['pytest-runner'],
    packages=find_packages(),
    install_requires=parse_requirements('requirements.txt')
)



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 12 Feb, 2018
#
from .exception import ServerException
from .pattern import NamedSingletonMixin
from tornado.options import define, options
from urllib.parse import urlparse

import aioredis
import asyncio
import functools
import logging

LOG = logging.getLogger('tornado.application')


class RedisConnectorError(ServerException):
    pass


class RedisConnector(NamedSingletonMixin):

    def __init__(self, name: str):
        self.name = name

    def connection(self):
        if not hasattr(self, '_connections') or not self._connections:
            raise RedisConnectorError("no connection of %s found" % self.name)
        return self._connections.get()

    async def connect(self, event_loop=None):
        name = self.name
        opts = options.group_dict('%s redis' % name)
        connection_string = opts[option_name(name, "uri")]
        r = urlparse(connection_string)
        if r.scheme.lower() != 'redis':
            raise RedisConnectorError('%s is not a redis connection scheme' %
                                      connection_string)
        num_connections = opts[option_name(name, "num-connections")]
        LOG.info('connecting redis [%s] %s' % (self.name, connection_string))
        if event_loop is None:
            event_loop = asyncio.get_event_loop()
        self._connections = await aioredis.create_pool(
            connection_string,
            encoding="UTF-8",
            minsize=int(num_connections[0]),
            maxsize=int(num_connections[-1]),
            loop=event_loop
        )


def option_name(instance: str, option: str) -> str:
    return 'redis-%s-%s' % (instance, option)


def register_redis_options(instance: str='master', default_uri: str='redis:///'):
    define(option_name(instance, "uri"),
           default=default_uri,
           group='%s redis' % instance,
           help="redis connection uri for %s" % instance)
    define(option_name(instance, 'num-connections'), multiple=True,
           default=[1, 2],
           group='%s redis' % instance,
           help='# of redis connections for %s ' % instance)


def with_redis(name: str):

    def wrapper(function):

        @functools.wraps(function)
        async def f(*args, **kwargs):
            async with RedisConnector.instance(name).connection() as redis:
                if "redis" in kwargs:
                    raise RedisConnectorError(
                        "duplicated database argument for redis %s" % name)
                kwargs.update({"redis": aioredis.Redis(redis)})
                retval = await function(*args, **kwargs)
                return retval
        return f

    return wrapper


def connect_redis(name: str):
    return RedisConnector.instance(name).connect



# -*- coding: utf-8 -*-
# The route helpers were originally written by
# Jeremy Kelley (http://github.com/nod).

# modified by jianing.yang (http://github.com/jianingy)

import tornado.web


class RouteFactory:
    _route_classes = dict()

    @classmethod
    def create(cls, route_name: str):
        route_cls = type('Route%s' % route_name.capitalize(),
                         (RouteManager,),
                         {"route_name": route_name})
        cls._route_classes[route_name] = route_cls
        return cls._route_classes[route_name]

    @classmethod
    def get_routes(cls, route_name: str):
        route_cls = cls._route_classes.get(route_name, None)
        if route_cls is None:
            return []
        return route_cls.get_routes()


class RouteManager(object):

    _routes = []

    def __init__(self, uri: str, name: str=None, redirect: str=None):
        self._uri = uri
        self._redirect = redirect
        self.name = name

    def __call__(self, _handler):
        """gets called when we class decorate"""
        name = self.name or _handler.__name__
        if self._redirect is not None:
            _handler = tornado.web.RedirectHandler
            self._routes.append(tornado.web.url(self._uri, _handler,
                                                dict(url=self._redirect),
                                                name=name))
        elif issubclass(_handler, tornado.web.StaticFileHandler):
            settings = {
                'path': _handler.path,
                'default_filename': _handler.default_filename,
            }
            self._routes.append(tornado.web.url(self._uri, _handler,
                                                name=name, kwargs=settings))
        else:
            self._routes.append(tornado.web.url(self._uri, _handler,
                                                name=name))
        return _handler

    @classmethod
    def get_routes(self):
        return self._routes


route = RouteFactory.create("default")



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @  8 Feb, 2018
#
from .exception import GeneralException, ClientException, ServerException
from ujson import loads as json_decode, dumps as json_encode
from tornado.options import options
from traceback import format_exception
import tornado.web


class InvalidJSONRequestData(ClientException):
    error_format = "Request data must be in JSON format"
    error_code = 400001


class JSONController(tornado.web.RequestHandler):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = dict()

    def prepare(self):
        super().prepare()
        content_type = self.request.headers.get('Content-Type')

        if not self.request.body:
            return

        if not content_type:
            return

        if content_type.strip().lower().startswith('application/json'):
            try:
                self.data = json_decode(self.request.body)
            except ValueError as e:
                raise InvalidJSONRequestData() from e

    def reply(self, *args, **kwargs):
        self.set_header('Content-Type', 'application/json; charset=UTF-8')
        self.finish(json_encode(kwargs))

    def bail(self, status_code: int, *args, **kwargs):
        self.set_header('Content-Type', 'application/json; charset=UTF-8')
        self.set_status(status_code)
        self.finish(json_encode(kwargs))

    def options(self, *args, **kwargs):
        if options.debug:
            self.set_header("Allow", "POST, GET, PUT, DELETE, OPTIONS, PATCH")

    def set_default_headers(self, *args, **kwargs):
        super().set_default_headers(*args, **kwargs)
        if options.debug:
            self.set_header("Access-Control-Allow-Origin", "*")
            self.set_header("Access-Control-Allow-Methods",
                            "POST, GET, PUT, DELETE, OPTIONS, PATCH")
            self.set_header("Access-Control-Max-Age", "3600")
            self.set_header("Access-Control-Allow-Headers",
                            "Content-Type, Access-Control-Allow-Headers")

    def get_data(self, name, default, strip=True):
        v = self.data.get(name, default)
        if strip and isinstance(v, str):
            v = v.strip()
        return v

    def write_error(self, status_code: int, **kwargs):
        if 'exc_info' in kwargs:
            retval = dict()
            exc_type, exc, trace = kwargs['exc_info']
            if isinstance(exc, GeneralException):
                retval["reason"] = exc.message
            else:
                retval["reason"] = str(exc)
            if isinstance(exc, ClientException):
                self.set_status(400)
            elif isinstance(exc, ServerException):
                self.set_status(500)
            if hasattr(exc, "http_status_code"):
                self.set_status(exc.http_status_code)
            if hasattr(exc, "error_code"):
                retval["status"] = exc.error_code
            else:
                retval["status"] = 500000
            if hasattr(options, 'debug') and options.debug:
                retval['traceback'] = format_exception(exc_type, exc, trace)
            self.reply(**retval)
        else:
            self.set_status(status_code)
            self.reply(reason="no detail information")



# -*- coding: utf-8 -*-
#
# This piece of code is written by
#    Jianing Yang <jianingy.yang@gmail.com>
# with love and passion!
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#                             22 Jan, 2016
#
from .exception import ServerException
from .pattern import NamedSingletonMixin
from tornado.options import define, options
from urllib.parse import urlparse

import asyncio
import aiopg
import functools
import logging

LOG = logging.getLogger('tornado.application')


class PostgresConnectorError(ServerException):
    pass


class PostgresConnector(NamedSingletonMixin):

    def __init__(self, name: str):
        self.name = name

    def setup_options(self):
        name = self.name
        opts = options.group_dict('%s database' % name)
        uri = opts[option_name(name, "uri")]
        r = urlparse(uri)
        if r.scheme.lower() != 'postgres':
            raise PostgresConnectorError('%s is not a postgres connection scheme' % uri)
        fmt = ('host={host} port={port} dbname={db} user={user} password={pw}')
        dsn = fmt.format(host=r.hostname or 'localhost',
                         port=r.port or 5432,
                         user=r.username,
                         pw=r.password,
                         db=r.path.lstrip('/') or r.username)
        self._connection_string = dsn
        self._reconnect_interval = opts[option_name(name, "reconnect-interval")]
        self._num_connections = opts[option_name(name, "num-connections")]

    def connection(self):
        if not hasattr(self, '_connections') or not self._connections:
            raise PostgresConnectorError("no connection of %s found" % self.name)
        return self._connections.acquire()

    async def connect(self, event_loop=None):
        self.setup_options()
        LOG.info('connecting postgresql [%s] %s' %
                 (self.name, self._connection_string))
        if event_loop is None:
            event_loop = asyncio.get_event_loop()
        self._connections = await aiopg.create_pool(
            dsn=self._connection_string,
            minsize=int(self._num_connections[0]),
            maxsize=int(self._num_connections[-1]),
            loop=event_loop,
        )
        return self._connections


def option_name(instance: str, option: str) -> str:
    return 'postgres-%s-%s' % (instance, option)


def register_postgres_options(instance: str='master', default_uri: str='postgres:///'):
    define(option_name(instance, "uri"),
           default=default_uri,
           group='%s database' % instance,
           help="postgresql connection uri for %s" % instance)
    define(option_name(instance, 'num-connections'), multiple=True,
           default=[1, 2],
           group='%s database' % instance,
           help='connection pool size for %s ' % instance)
    define(option_name(instance, 'reconnect-interval'),
           default=5,
           group='%s database' % instance,
           help='reconnect interval for %s' % instance)


def with_postgres(name: str):

    def wrapper(function):

        @functools.wraps(function)
        async def f(*args, **kwargs):
            async with PostgresConnector.instance(name).connection() as db:
                if "db" in kwargs:
                    raise PostgresConnectorError(
                        "duplicated database argument for database %s" % name)
                kwargs.update({"db": db})
                retval = await function(*args, **kwargs)
                return retval
        return f

    return wrapper


def connect_postgres(name: str):
    return PostgresConnector.instance(name).connect



# -*- coding: utf-8 -*-
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#               Jianing Yang @ 16 Feb, 2018
#
from tornado.options import options, define
from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application as WebApplication
from tornado_battery.controller import JSONController
from tornado_battery.exception import ClientException, ServerException
from ujson import dumps as json_encode
from unittest import mock

define("debug", group="main", default=False, help="enable debug")

class SimpleController(JSONController):

    def get(self):
        return self.reply(quote="may the force be with you")

    async def post(self):
        return self.reply(data=self.data)

    def put(self):
        return self.reply(quote=self.get_data("quote", "no data"))

    def patch(self):
        return self.reply(quote=self.get_data("quote", "no data", strip=False))


class BailController(JSONController):

    def get(self):
        self.bail(503, reason="something wrong")


class ClientExceptionController(JSONController):

    def get(self):
        raise ClientException(reason="something wrong")


class ServerExceptionController(JSONController):

    def get(self):
        raise ServerException(reason="something wrong")


class PythonExceptionController(JSONController):

    def get(self):
        raise ValueError("a value error")


class ExceptionWithCode(ClientException):
    http_status_code = 403


class UnauthorizedController(JSONController):

    def get(self):
        raise ExceptionWithCode()

class TestController(AsyncHTTPTestCase):

    def get_app(self):
        app = WebApplication([
            (r"/", SimpleController),
            (r"/bail", BailController),
            (r"/exception/client", ClientExceptionController),
            (r"/exception/server", ServerExceptionController),
            (r"/exception/python", PythonExceptionController),
            (r"/exception/403", UnauthorizedController),
        ])
        return app

    def test_json_reply(self):
        response = self.fetch("/")
        assert response.body == b'{"quote":"may the force be with you"}'
        assert response.headers["Content-Type"] == "application/json; charset=UTF-8"

    def test_bail(self):
        response = self.fetch("/bail")
        assert response.code == 503
        assert response.body == b'{"reason":"something wrong"}'

    def test_json_request(self):
        data = json_encode(dict(x=1, y=2))
        headers = {
            "Content-Type": "application/json"
        }
        response = self.fetch("/", method="POST", body=data, headers=headers)
        assert response.body == b'{"data":{"x":1,"y":2}}'

    def test_get_data(self):
        data = json_encode(dict(quote="hello, world"))
        headers = {
            "Content-Type": "application/json"
        }
        response = self.fetch("/", method="PUT", body=data, headers=headers)
        assert response.body == b'{"quote":"hello, world"}'

    def test_get_data_num(self):
        data = json_encode(dict(quote=1984))
        headers = {
            "Content-Type": "application/json"
        }
        response = self.fetch("/", method="PUT", body=data, headers=headers)
        assert response.body == b'{"quote":1984}'

    def test_get_data_empty(self):
        data = json_encode(dict(non="hello, world"))
        headers = {
            "Content-Type": "application/json"
        }
        response = self.fetch("/", method="PUT", body=data, headers=headers)
        assert response.body == b'{"quote":"no data"}'

    def test_get_data_strip(self):
        data = json_encode(dict(quote="\t\r\n vvv  \r\t\n  "))
        headers = {
            "Content-Type": "application/json"
        }
        response = self.fetch("/", method="PUT", body=data, headers=headers)
        assert response.body == b'{"quote":"vvv"}'

    def test_get_data_dont_strip(self):
        data = json_encode(dict(quote="\t\r\n vvv  \r\t\n  "))
        headers = {
            "Content-Type": "application/json"
        }
        response = self.fetch("/", method="PATCH", body=data, headers=headers)
        assert response.body == b'{"quote":"\\t\\r\\n vvv  \\r\\t\\n  "}'

    def test_json_request_no_content_type(self):
        data = json_encode(dict(x=1, y=2))
        headers = {
            "Content-Type": ""
        }
        response = self.fetch("/", method="POST", body=data, headers=headers)
        assert response.body == b'{"data":{}}'

    def test_json_request_invalid_json_data(self):
        data = "{Invalid JSON}"
        headers = {
            "Content-Type": "application/json"
        }
        response = self.fetch("/", method="POST", body=data, headers=headers)
        match = (b'{"reason":"Request data must be in JSON format","status":400001}')
        assert response.body == match

    def test_client_exception(self):
        response = self.fetch("/exception/client")
        match = b'{"reason":"something wrong","status":400000}'
        assert response.body == match

    def test_server_exception(self):
        response = self.fetch("/exception/server")
        match = b'{"reason":"something wrong","status":500000}'
        assert response.body == match

    def test_python_exception(self):
        response = self.fetch("/exception/python")
        match = b'{"reason":"a value error","status":500000}'
        assert response.body == match

    def test_http_code(self):
        response = self.fetch("/exception/403")
        assert response.code == 403

    def test_allow_origin(self):
        response = self.fetch("/")
        assert "Access-Control-Allow-Origin" not in response.headers

        with mock.patch.object(options.mockable(), 'debug', True):
            response = self.fetch("/")
            assert response.headers["Access-Control-Allow-Origin"] == "*"
